#!/usr/bin/env python
# -*- coding: utf-8 -*-
#*****************************************************************************
#    Copyright (C) 2006-2007 Serge Weinstock <serge.weinstock@gmail.com>
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************
from qt import * #@UnusedWildImport
from kdecore import * #@UnusedWildImport
from kdeui import * #@UnusedWildImport
from kio import * #@UnusedWildImport
from RegexRenameDlg import RegexRenameDlg

NMAX_HISTORY = 10

class FRregex(object):
    """A wrapper for the regular expression operations"""
    def __init__(self):
        """Constructor"""
        self.re = QRegExp()
        self.result = []
    def setRE(self, re = None, caseInsensitive = None):
        """
        set the regular expression

        @param[in] re the regular expression
        @param[in] caseInsensitive if True, the regular expression is case insensitive
        """
        if re is not None:
            self.re.setPattern(re)
        if caseInsensitive is not None:
            self.re.setCaseSensitive(not caseInsensitive)
    def setResult(self, result):
        """
        set the expression used for generating tranformed strings

        @param[in] result the expression
        """
        self.result = []
        opos = 0
        spos = 0
        while True:
            pos = result.find('\\', spos)
            if (pos < 0) or (pos == len(result) - 1):
                if self.result and isinstance(self.result[-1], QString):
                    self.result[-1] += result[opos:]
                else:
                    self.result.append(result[opos:])
                break
            if (result[pos + 1] == '\\'):
                if self.result and isinstance(self.result[-1], QString):
                    self.result[-1] += result[opos:pos+1]
                else:
                    self.result.append(result[opos:pos+1])
                opos = pos + 2
                spos = opos
                continue
            group, isdigit = result[pos + 1].toInt()
            if (isdigit):
                self.result.append(result[opos:pos])
                self.result.append(group)
                opos = pos + 2
                spos = opos
                continue
            spos = pos + 1
    def apply(self, text):
        """
        Apply the regular expression to text

        @param[in] text the text to process
        @return the transformed text or None if the operation failed
        """
        if (not self.result) or (not self.re.exactMatch(text)):
            return None
        res = QString()
        for elt in self.result:
            if isinstance(elt, QString):
                res += elt
            else:
                if elt > self.re.numCaptures():
                    return None
                res += self.re.cap(elt)
        return res

class URLListItem(QCheckListItem):
    """A URL as shown in the main list view of the dialog"""
    STATUS_OK = 0
    STATUS_WARNING = 1
    STATUS_DUPLICATED = 2
    STATUS_ERROR = 3
    def __init__(self, url, parent, owner):
        '''
        Constructor

        @param[in] url the item URL
        @param[in] parent the parent listview item
        @param[in] owner the owner UI component
        '''
        super(URLListItem, self).__init__(parent, "", QCheckListItem.CheckBox)
        assert(owner)
        self.owner = owner
        self.status = URLListItem.STATUS_OK
        self.url = KURL(url)
        self.newURL = None
        self.renamable = self.url.hasPath()
        if not self.renamable:
            self.setText(0, self.url.prettyURL())
            self.setEnabled(False)
            return
        self.setText(0, self.url.fileName())
        self.setOn(True)
    def canRename(self):
        """Return 1 if the item can be renamed"""
        if (self.status == URLListItem.STATUS_OK) and self.isOn() and self.renamable:
            return 1
        return 0
    def showError(self):
        """
        Display this element as an error

        @return 0 if the element is checked 1 otherwise
        """
        self.setText(1, i18n('<error>'))
        if self.isOn():
            self.status = URLListItem.STATUS_ERROR
            return 0
        self.status = URLListItem.STATUS_WARNING
        return 1
    def setAsDuplicated(self, text):
        """
        Display this element as duplicated

        @param[in] text the text to show as the new name
        """
        self.status = URLListItem.STATUS_DUPLICATED
        self.setText(1, text)
    def applyRE(self, re):
        """
        Apply a FRregex to this element

        @param[in] re the FRegex
        @return (status, name) where
            status equals 1 if successfull 0 otherwise
            name is the new file name if the conversion was successfull and the item will be renamed
        """
        if not self.renamable:
            return (1, None)
        res = re.apply(self.url.fileName())
        if res is None:
            return (self.showError(), None)
        res.stripWhiteSpace()
        if res.isEmpty():
            return (self.showError(), None)
        self.setText(1, res)
        self.status = URLListItem.STATUS_OK
        if self.isOn():
            return (1, res)
        return (1, None)
    def paintCell(self, painter, colorGroup, column, width, alignment):
        """
        Paint a cell
        It paint the text
        - in the normal color if the tranformation was successfull
        - in red if the tranformation was unsuccessfull and the item is checked
        - in dark magenta if the tranformation was unsuccessfull and the item is not checked

        @param[in] painter  the QPainter open on the relevant paint device
        @param[in] colorGroup  the color group to use
        @param[in] column  the logical column number within the item that is to be painted
        @param[in] width the cell width
        @param[in] alignment the cell alignment
        """
        _cg = QColorGroup(colorGroup)
        c = _cg.text();
        if self.status == URLListItem.STATUS_ERROR:
            _cg.setColor(QColorGroup.Text, Qt.red)
        elif self.status == URLListItem.STATUS_DUPLICATED:
            _cg.setColor(QColorGroup.Text, Qt.darkMagenta)
        elif self.status == URLListItem.STATUS_WARNING:
            _cg.setColor(QColorGroup.Text, Qt.darkBlue)
        QCheckListItem.paintCell(self, painter, _cg, column, width, alignment)
        _cg.setColor(QColorGroup.Text, c)
    def stateChange(self, newState): #@UnusedVariable
        '''
        This is called when the user clicks on the checkbox

        @param[in] newState set to True if the item is now checked
        '''
        # notify the parent
        self.owner.onItemStatusChange()
    def doIt(self):
        """
        Do the rename operation
        @return True if the operation was successfull
        """
        self.newURL = None
        if self.canRename():
            newURL = KURL(self.url)
            newURL.setFileName(self.text(1))
            KIO.NetAccess.move(self.url, newURL, self.owner)
            if KIO.NetAccess.lastError():
                self.status == URLListItem.STATUS_ERROR
                self.setText(2, KIO.NetAccess.lastErrorString())
                return False
            self.status == URLListItem.STATUS_OK
            self.setText(2, i18n("OK"))
            self.newURL = newURL
            return True
    def undo(self):
        """
        Restore the item to its original state
        @return True if the operation was successfull
        """
        if not self.newURL:
            # the item has not been renamed
            return True
        KIO.NetAccess.move(self.newURL, self.url, self.owner)
        if KIO.NetAccess.lastError():
            self.status == URLListItem.STATUS_ERROR
            self.setText(2, KIO.NetAccess.lastErrorString())
            self.url = self.newURL
            self.newURL = None
            return False
        self.newURL = None
        return True
    def copyFilename(self):
        """Copy this item filename to the clipboard"""
        KApplication.kApplication().clipboard().setText(self.url.fileName())

class RegexRenameDlgImpl(RegexRenameDlg):
    """This dialog allows to rename files with the help of regular expressions"""
    def __init__(self, options, parent = None,name = None,modal = 0,fl = 0):
        '''
        Constructor
        @param[in] options the parameters read from the command line arguments
        @param[in] parent the dialog parent
        @param[in] name the dialog name
        @param[in] modal True if the dialog is modal
        @param[in] fl the dialog contruction flags
        '''
        RegexRenameDlg.__init__(self,parent,name,modal,fl)
        #
        self.afterRename = False
        self.re = FRregex()
        # fill up the list
        self.N = 0
        for i in xrange(1, options.count()):
            URLListItem(options.url(i), self.lvResults, self)
            self.N += 1
        self.__reset()
        self.connect(self.lvResults, SIGNAL("contextMenu(KListView*,QListViewItem*,const QPoint&)"), self.onLvResultsContextMenu)
        # set up the dialog
        cfg = KApplication.kApplication().config()
        cfg.setGroup("RegexRename")
        self.chkCaseInsensitive.setChecked(cfg.readBoolEntry("caseInsensitive", True))
        reList = cfg.readListEntry("re_in")
        if reList: self.txtREin.setText(reList[-1])
        self.txtREin.completionObject().setItems(reList)
        reList = cfg.readListEntry("re_out")
        if reList: self.txtREout.setText(reList[-1])
        self.txtREout.completionObject().setItems(reList)
        #
        self.buttonOk.setIconSet(SmallIconSet("ok"))
        self.buttonCancel.setIconSet(SmallIconSet("cancel"))
        self.buttonHelp.setIconSet(SmallIconSet("help"))
        #
        w = cfg.readUnsignedNumEntry("width")
        h = cfg.readUnsignedNumEntry("height")
        if w > 0 and h > 0:
            self.resize(w, h)
    def __reset(self):
        """Reset the listview elements by applying a new re"""
        # let first apply the regex
        self.lvResults.setUpdatesEnabled(False)
        li = self.lvResults.firstChild()
        nOK = 0
        new_names = {}
        while li:
            OK, new_name = li.applyRE(self.re)
            nOK += OK
            if OK and new_name:
                new_names.setdefault(unicode(new_name), []).append(li)
            li = li.nextSibling()
        # now look for duplicates
        fmt = i18n("%1 <%2 files with the same name>")
        for new_name, dups in new_names.iteritems():
            if len(dups) > 1:
                msg = fmt.arg(new_name).arg(len(dups))
                for item in dups:
                    item.setAsDuplicated(msg)
                    nOK -= 1
        #
        self.lvResults.setUpdatesEnabled(True)
        self.lvResults.triggerUpdate()
        self.buttonOk.setEnabled(bool((nOK == self.N) and (new_names)))
    def onItemStatusChange(self):
        """
        Called when the state of one of the list view items has changed
        """
        self.__reset()
    def onREinChange(self,a0):
        """
        Called when the content of the REin textbox has changed
        @param[in] a0 the textbox contents
        """
        self.re.setRE(re = a0)
        self.__reset()
    def onREoutChange(self,a0):
        """
        Called when the content of the REout textbox has changed
        @param[in] a0 the textbox contents
        """
        self.re.setResult(a0)
        self.__reset()
    def onCaseInsensitiveToggle(self, a0):
        """
        Called when the status of the CaseInsensitive checkbox has changed
        @param[in] a0 the checkbox status
        """
        self.re.setRE(caseInsensitive = a0)
        self.__reset()
    def onRenameOrUndo(self):
        """
        Called when the user clicks on the "Rename" or "Undo" button
        """
        if self.afterRename:
            self.doUndo()
        else:
            self.doRename()
    def doRename(self):
        """
        Called when the user clicks on the "Rename" button
        """
        # save the prefs
        cfg = KApplication.kApplication().config()
        cfg.setGroup("RegexRename")
        cfg.writeEntry("caseInsensitive", self.chkCaseInsensitive.isChecked())
        #
        reList = cfg.readListEntry("re_in")
        txt = self.txtREin.text()
        while reList.contains(txt):
            reList.remove(txt)
        if len(reList) >= NMAX_HISTORY-1:
            del reList[0:len(reList)-NMAX_HISTORY+1]
        reList.append(txt)
        cfg.writeEntry("re_in", reList)
        #
        reList = cfg.readListEntry("re_out")
        txt = self.txtREout.text()
        while reList.contains(txt):
            reList.remove(txt)
        if len(reList) >= NMAX_HISTORY-1:
            del reList[0:len(reList)-NMAX_HISTORY+1]
        reList.append(txt)
        cfg.writeEntry("re_out", reList)
        # set the UI
        self.buttonCancel.setText(i18n("&Close"))
        self.buttonOk.setText(i18n("&Undo"))
        self.lvResults.addColumn(i18n("Results"), -1)
        for i in xrange(self.lvResults.columns()):
            self.lvResults.adjustColumn(i)
        for control in [self.txtREin, self.txtREout, self.chkCaseInsensitive]:
            control.setEnabled(False)
        li = self.lvResults.firstChild()
        while li:
            li.setEnabled(False)
            li = li.nextSibling()
        # let's go
        li = self.lvResults.firstChild()
        while li:
            li.doIt()
            li = li.nextSibling()
        self.afterRename = True
    def doUndo(self):
        """
        Called when the user clicks on the "undo" button
        """
        pbs = []
        li = self.lvResults.firstChild()
        while li:
            if not li.undo():
                pbs.append(li)
            li = li.nextSibling()
        if pbs:
            lpbs = QStringList()
            for pb in pbs:
                lpbs.append(pb.url.filename())
            KMessageBox.errorList(
                    None,
                    i18n("""Unable to rename %1 file(s)""").arg(len(pbs)),
                    lpbs
                    )
        li = self.lvResults.firstChild()
        while li:
            li.setEnabled(True)
            li = li.nextSibling()
        self.buttonOk.setText(i18n("&Rename"))
        self.buttonCancel.setText(i18n("&Cancel"))
        self.lvResults.removeColumn(self.lvResults.columns()-1)
        self.afterRename = False
    def onCancel(self):
        """Close the dialog"""
        # save the prefs
        cfg = KApplication.kApplication().config()
        cfg.setGroup("RegexRename")
        cfg.writeEntry("width", self.width())
        cfg.writeEntry("height", self.height())
        # banzai!
        self.reject()
    def onLvResultsContextMenu(self, lv, item, point): #@UnusedVariable
        """
        Callback for the list view context menu
        """
        cmenu = KPopupMenu(self,"MyActions")
        cmenu.insertItem(i18n("Copy file name"), item.copyFilename)
        cmenu.exec_loop(point)
    def clean(self):
        """
        Removes circular references that can lead to a crash when the dialog is destroyed
        """
        self.lvResults.clear()

def run(options):
    """
    Create and launch the dialog and do the associated actions.

    param[in] options the program options
    """
    kapp = KApplication.kApplication()
    dialog = RegexRenameDlgImpl(options)
    kapp.setMainWidget(dialog)
    dialog.show()
    res = kapp.exec_loop()
    dialog.clean()
    return res
