#!/usr/bin/env python
# -*- coding: utf-8 -*-
#*****************************************************************************
#    Copyright (C) 2006-2007 Serge Weinstock <serge.weinstock@gmail.com>
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#*****************************************************************************
from qt import * #@UnusedWildImport
from kdecore import * #@UnusedWildImport
from kdeui import * #@UnusedWildImport
from kio import * #@UnusedWildImport
from kdeutils import DirListerEx
from CleanDlg import CleanDlg

DEF_GROUPS = [
              (i18n("Backup files"), ["*~", "*.bak", "*.BAK"]),
              (i18n("Object files"), ["*.o", "*.ko", "*.lo", "*.pyc", "*.gmo"]),
              (i18n("Libraries files"), ["*.a", "*.so", "*.so.*", "*.la"]),
              (i18n("Build directories"), ["Debug", "Release", ".objs", ".libs", ".deps"]),
              ]

def formatSize(sz):
    """
    Returns a string containing a human readable version of a file size
    """
    if sz < 0x400:
        return i18n("%1  ").arg(float(sz), 8, 'f',2)
    if sz < 0x100000:
        return i18n("%1 Kb").arg(sz / float(0x400), 8, 'f', 2)
    if sz < 0x40000000:
        return i18n("%1 Mb").arg(sz / float(0x100000), 8, 'f', 2)
    return i18n("%1 Gb").arg(sz / float(0x40000000), 8, 'f', 2)

class GroupItem(QCheckListItem):
    COL_NAME = 0
    COL_SIZE = 1
    COL_NITEMS = 2
    """
    A group as shown in the list view
    """
    def __init__(self, parent, owner, after = None, cfg = None, igroup = 0, name = "", patterns = [], on = True):
        """
        Constructor
        @param[in] parent the item parent
        @param[in] owner the owner UI component
        @param[in] after the item to insert after
        @param[in] cfg the application config
        @param[in] igroup the group number in the config
        @param[in] name the entry name
        @param[in] patterns the patterns
        @param[in] on group on?
        """
        assert(owner)
        self.owner = owner
        self.nitems = 0
        self.sz = 0
        # first parse the config to see if the item is valid
        if cfg:
            name = cfg.readEntry("group%d_Name" % igroup)
            patterns = [unicode(x) for x in cfg.readListEntry("group%d_Patterns" % igroup)]
            on = cfg.readBoolEntry("group%d_On" % igroup, True)
        if not name: raise ValueError()
        # then add it to the list view
        super(GroupItem, self).__init__(parent, after, "", QCheckListItem.CheckBox)
        self.patterns = [ QRegExp(p, True, True) for p in patterns ]
        self.setText(self.COL_NAME, name)
        self.setOn(on)
    def stateChange(self, newState): #@UnusedVariable
        '''
        This is called when the user clicks on the checkbox

        @param[in] newState set to True if the item is now checked
        '''
        # notify the parent
        self.initOp()
        self.owner.onItemStatusChange()
    def setPatternsFromString(self, str):
        """
        Set the patterm from a ',' separated list of patterns
        """
        self.patterns = [ QRegExp(x.strip(), True, True) for x in unicode(str).split(u",")]
    def stringFromPatterns(self):
        """
        Returns a ',' separated list of patterns
        """
        return u','.join([unicode(p.pattern()) for p in self.patterns])
    def save(self, cfg, igroup):
        """
        Constructor
        @param[in] cfg the application config
        @param[in] igroup the group number in the config
        """
        cfg.writeEntry("group%d_Name" % igroup, self.text(0))
        cfg.writeEntry("group%d_Patterns" % igroup, [unicode(p.pattern()) for p in self.patterns])
        cfg.writeEntry("group%d_On" % igroup, self.isOn())
    def initOp(self):
        """
        Initialization of a search or delete operation
        """
        self.nitems = 0
        self.sz = 0
        self.setText(self.COL_SIZE, formatSize(self.sz))
        self.setText(self.COL_NITEMS, str(self.nitems))
    def iterOp(self, files):
        """
        Iteration of a search or delete operation
        @parm files a list of KFileItem to process
        """
        if (not self.isOn()) or (not files):
            return
        # let first see if this folder or a parent folder is deleted
        delAll = False
        for d in unicode(files[0].url().upURL().path()).split(u'/'):
            for patt in self.patterns:
                if patt.exactMatch(d):
                    delAll = True
                    break
        # now let's process all the files and subfolders
        for f in files: #@UnusedVariable
            if not hasattr(f, "toDelete"): # not already marked in another group
                if f.isWritable(): # we must be able to delete it
                    if f.isDir() or f.isFile():
                        if delAll:
                            # delete everything in this folder
                            f.toDelete = True
                            self.nitems += 1
                            self.sz += f.size()
                        else:
                            # match on the file name ?
                            path = f.url().filename()
                            for patt in self.patterns:
                                if patt.exactMatch(path):
                                    f.toDelete = True
                                    self.nitems += 1
                                    self.sz += f.size()
                                    break
        self.setText(self.COL_SIZE, formatSize(self.sz))
        self.setText(self.COL_NITEMS, str(self.nitems))
    def paintCell(self, painter, colorGroup, column, width, alignment):
        """
        Paint a cell
        It paint the text
        - in the normal color if the tranformation was successfull
        - in red if the tranformation was unsuccessfull and the item is checked
        - in dark magenta if the tranformation was unsuccessfull and the item is not checked

        @param[in] painter  the QPainter open on the relevant paint device
        @param[in] colorGroup  the color group to use
        @param[in] column  the logical column number within the item that is to be painted
        @param[in] width the cell width
        @param[in] alignment the cell alignment
        """
        _cg = QColorGroup(colorGroup)
        c = _cg.text();
        if self.owner.deleting:
            _cg.setColor(QColorGroup.Text, Qt.darkGray)
        QCheckListItem.paintCell(self, painter, _cg, column, width, alignment);
        _cg.setColor(QColorGroup.Text, c)

class CleanDlgImpl(CleanDlg):
    """This dialog allows to rename files with the help of regular expressions"""
    ACTION_NONE = 0
    ACTION_LIST = 1
    ACTION_DEL  = 2
    def __init__(self, options, parent = None,name = None,modal = 0,fl = 0):
        '''
        Constructor
        @param[in] options the parameters read from the command line arguments
        @param[in] parent the dialog parent
        @param[in] name the dialog name
        @param[in] modal True if the dialog is modal
        @param[in] fl the dialog contruction flags
        '''
        CleanDlg.__init__(self,parent,name,modal,fl)
        #
        self.dirLister = None
        self.action = self.ACTION_NONE
        self.deleting = False
        self.foldersToDelete = []
        # check the arguments
        if options.count() < 2:
            raise RuntimeError(i18n("Invalid number of arguments"))
        # folders to process and title
        self.folders = []
        title = i18n("Cleaning ")
        for i in xrange(1, options.count()):
            d = KURL(options.url(i))
            self.folders.append(d)
            if i > 1:
                title += ", "
            title += d.fileName()
        self.setCaption(title)
        # set up the dialog
        self.groups = []
        cfg = KApplication.kApplication().config()
        cfg.setGroup("CleanDlg")
        nCfgs = cfg.readUnsignedNumEntry("nGroups")
        for i in xrange(nCfgs):
            try:
                GroupItem(self.listGroups, self, cfg = cfg, igroup = i)
            except ValueError:
                break
        if not self.listGroups.childCount():
            for name, patterns in DEF_GROUPS:
                GroupItem(self.listGroups, self, name = name, patterns = patterns)
        self.listGroups.setSelected(self.listGroups.firstChild(), True)
        #
        self.buttonOk.setIconSet(SmallIconSet("ok"))
        self.buttonCancel.setIconSet(SmallIconSet("cancel"))
        self.buttonHelp.setIconSet(SmallIconSet("help"))
        #
        self.progress.setTotalSteps(100)
        self.progress.setProgress(100)
        #
        self.layout().activate() # ensure that all widgets are correctly sized
        self.expandDialog(False) # hide the edit group panel
        #
        w = cfg.readUnsignedNumEntry("width")
        h = cfg.readUnsignedNumEntry("height")
        if w > 0 and h > 0:
            self.resize(w, h)
        # don't launch a search immediatly, let all the previous events to be processed first
        # thus the dialog will pop up quicker
        QTimer.singleShot(1, self.__onTimer)
    def __onTimer(self):
        # let's go
        self.__startDirLister(self.ACTION_LIST)
    def expandDialog(self, expand):
        """use any picture as a folder icon
        Handle expansion/shriking of the dialog
        """
        sz = self.size()
        if expand:
            self.frameGroups.show()
            sz.setHeight(sz.height()+self.frameGroups.height()+2*self.layout().spacing())
            self.buttonEditGroups.setIconSet(SmallIconSet("up"))
            self.buttonEditGroups.setText(i18n("Edit &Groups <<"))
            self.expanded = True
        else:
            sz.setHeight(sz.height()-self.frameGroups.height()-2*self.layout().spacing())
            self.frameGroups.hide()
            self.buttonEditGroups.setIconSet(SmallIconSet("down"))
            self.buttonEditGroups.setText(i18n("Edit &Groups >>"))
            self.expanded = False
        self.layout().activate()
        self.resize(sz)
    def __update(self, launchNewSearch = False):
        """
        Update the enable/disable status of the widgets
        @param[in] launchNewSearch if True, launch also a new search
        """
        hasSel = bool(self.listGroups.selectedItem())
        hasName = not self.lineGroupName.text().stripWhiteSpace().isEmpty()
        hasSthg = False
        li = self.listGroups.firstChild()
        while li:
            if li.isOn():
                hasSthg = True
                break
            li = li.nextSibling()
        self.buttonDelete.setEnabled((not self.deleting) and hasSel)
        self.buttonModify.setEnabled((not self.deleting) and hasSel and hasName)
        self.buttonAdd.setEnabled((not self.deleting) and hasName)
        self.buttonOk.setEnabled((not self.deleting) and hasSthg)
        #self.listGroups.setEnabled(not self.deleting)
        self.lineGroupName.setEnabled(not self.deleting)
        self.lineGroupPatterns.setEnabled(not self.deleting)
        if hasSthg and launchNewSearch:
            # relaunch a search
            self.__startDirLister(self.ACTION_LIST)
    def onEditGroups(self):
        """
        Callback for the Edit Groups button
        """
        self.expandDialog(not self.frameGroups.isVisible())
    def onGroupSel(self):
        """
        Callback for the group selection changes
        """
        li = self.listGroups.selectedItem()
        if li:
            self.lineGroupName.setText(li.text(0))
            self.lineGroupPatterns.setText(li.stringFromPatterns())
        self.__update()
    def onModifyGroup(self):
        """
        Callback for the Modify Group button
        """
        li = self.listGroups.selectedItem()
        if li:
            name = self.lineGroupName.text().stripWhiteSpace()
            if not name.isEmpty():
                li.setText(0, name)
                li.setPatternsFromString(self.lineGroupPatterns.text())

    def onAddGroup(self):
        """
        Callback for the Add Group button
        """
        name = self.lineGroupName.text().stripWhiteSpace()
        if not name.isEmpty():
            li = GroupItem(self.listGroups, self, after = self.listGroups.selectedItem(), name = name, patterns = [])
            li.setPatternsFromString(self.lineGroupPatterns.text())
            self.listGroups.setSelected(li, True)
            self.__update(True)
    def onDeleteGroup(self):
        """
        Callback for the Delete Group button
        """
        li = self.listGroups.selectedItem()
        if li:
            nsel = li.nextSibling()
            if not nsel: nsel = li.itemAbove()
            self.listGroups.takeItem(li)
            if nsel: self.listGroups.setSelected(nsel, True)
            self.__update(True)
    def onGroupNameChanged(self,a0): #@UnusedVariable
        """
        Called when the name of the group changes
        """
        self.__update()
    def onItemStatusChange(self):
        """
        Called when the state of one of the list view items has changed
        """
        if not self.deleting:
            self.__update(True)
    def hideEvent(self, e): #@UnusedVariable
        """
        Callback hacked for detecting when the dialog is closed by any mean
        """
        # if cancel has been clicked, let's also stop any job
        self.__stopDirLister()
        # let's save the config
        cfg = KApplication.kApplication().config()
        cfg.setGroup("CleanDlg")
        cfg.writeEntry("nGroups", self.listGroups.childCount())
        li = self.listGroups.firstChild()
        i = 0
        while li:
            li.save(cfg, i)
            li = li.nextSibling()
            i += 1
        sz = self.size()
        if self.expanded:
            sz.setHeight(sz.height()-self.frameGroups.height()-2*self.layout().spacing())
        cfg.writeEntry("width", sz.width())
        cfg.writeEntry("height", sz.height())
    def __startDirLister(self, action):
        """
        Start a new use of the dir lister
        """
        self.__stopDirLister()
        self.action = action
        self.dirLister = DirListerEx()
        self.connect(self.dirLister,PYSIGNAL("newItems"), self.newItems)
        self.connect(self.dirLister,PYSIGNAL("completed"), self.completed)
        self.progress.setTotalSteps(0)
        self.progress.reset()
        li = self.listGroups.firstChild()
        while li:
            li.initOp()
            li = li.nextSibling()
        self.dirLister.openURLs(self.folders, True, win = self)
    def __stopDirLister(self):
        """
        Stop any current use of the dir lister
        """
        if self.dirLister:
            self.disconnect(self.dirLister,PYSIGNAL("newItems"), self.newItems)
            self.disconnect(self.dirLister,PYSIGNAL("completed"), self.completed)
            self.dirLister.stop()
            self.dirLister = None
        self.action = self.ACTION_NONE
        self.progress.setTotalSteps(100)
        self.progress.setProgress(100)
    def clean(self):
        """
        Removes circular references that can lead to a crash when the dialog is destroyed
        """
        self.listGroups.clear()
    def newItems(self, items):
        '''
        Slot for kdeutils.DirLister.newItems
        @param[in] items the list of found items
        '''
        self.progress.setProgress(self.progress.progress()+1)
        li = self.listGroups.firstChild()
        while li:
            li.iterOp(items)
            li = li.nextSibling()
        if self.action == self.ACTION_DEL:
            for f in items:
                if hasattr(f, "toDelete"):
                    if f.isDir():
                        self.foldersToDelete.append(KURL(f.url()))
                    else:
                        KIO.file_delete(f.url(), False)
    def completed(self, success): #@UnusedVariable
        '''
        Slot for kdeutils.DirLister.completed
        @param[in] success true if the listing was successfull
        '''
        self.progress.setTotalSteps(100)
        self.progress.setProgress(100)
        if self.deleting:
            for d in self.foldersToDelete:
                KIO.del_(d, False, False)
            self.buttonCancel.setIconSet(SmallIconSet("exit"))
            self.buttonCancel.setText(i18n("&Close"))
    def onDeleteFiles(self):
        """
        Launch the file deletion
        """
        self.deleting = True
        self.foldersToDelete = []
        self.__update()
        self.__startDirLister(self.ACTION_DEL)

def run(options):
    """
    Create and launch the dialog and do the associated actions.

    param[in] options the program options
    """
    kapp = KApplication.kApplication()
    dialog = CleanDlgImpl(options)
    kapp.setMainWidget(dialog)
    dialog.show()
    res = kapp.exec_loop()
    dialog.clean()
    return res
